# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import numpy as np


def _is_array_like(a):
    """
    Helper function to determine if a value is array like.

    a : obj
        Object to test.

    Returns
    -------
    True or false respectively.
    """
    return isinstance(a, (list, tuple, np.ndarray))


def _is_one_dimensional(a):
    """
    Helper function to determine if value is one dimensional.

    a : array_like
        Object to test.

    Returns
    -------
    True or false respectively.
    """
    return a.ndim == 1


def _to_np_array(a):
    """
    Helper function to convert tuple or list to np.ndarray.

    a : Tuple, list or np.ndarray
        The object to transform.

    Returns
    -------
    The np.ndarray.

    Raises
    ------
    ValueError
        If a is not a valid type.
    """
    if not _is_array_like(a):
        raise ValueError('Unable to convert to np.ndarray!')

    return np.array(a)


def _rolling_window(a, window):
    """
    Provides a rolling window on a numpy array given an array and window size.

    Parameters
    ----------
    a : array_like
        The array to create a rolling window on.
    window : int
        The window size.

    Returns
    -------
    Strided array for computation.
    """
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


def _moving_average(a, window=3):
    """
    Computes the moving average over an array given a window size.

    Parameters
    ----------
    a : array_like
        The array to compute the moving average on.
    window : int
        The window size.

    Returns
    -------
    The moving average over the array.
    """
    return np.mean(_rolling_window(a, window), -1)


def _moving_std(a, window=3):
    """
    Computes the moving std. over an array given a window size.

    Parameters
    ----------
    a : array_like
        The array to compute the moving std. on.
    window : int
        The window size.

    Returns
    -------
    The moving std. over the array.
    """
    return np.std(_rolling_window(a, window), -1)


def _precheck_series_and_query(ts, query):
    """
    Helper function to ensure we have 1d time series and query.

    Parameters
    ----------
    ts : array_like
        The array to create a rolling window on.
    query : array_like
        The query.

    Returns
    -------
    (np.array, np.array) - The ts and query respectively.

    Raises
    ------
    ValueError
        If ts is not a list or np.array.
        If query is not a list or np.array.
        If ts or query is not one dimensional.
    """
    try:
        ts = _to_np_array(ts)
    except ValueError:
        raise ValueError('Invalid ts value given. Must be array_like!')

    try:
        query = _to_np_array(query)
    except ValueError:
        raise ValueError('Invalid query value given. Must be array_like!')

    if not _is_one_dimensional(ts):
        raise ValueError('ts must be one dimensional!')

    if not _is_one_dimensional(query):
        raise ValueError('query must be one dimensional!')

    return (ts, query)


def mass(ts, query, normalize_query=True, corr_coef=False):
    """
    Compute the distance profile for the given query over the given time 
    series. Optionally, the correlation coefficient can be returned and 
    the query can be normalized.

    Parameters
    ----------
    ts : array_like
        The array to create a rolling window on.
    query : array_like
        The query.
    normalize_query : bool, default True
        Optionally normalize the query.
    corr_coef : bool, default False
        Optionally return the correlation coef.

    Returns
    -------
    An array of distances.

    Raises
    ------
    ValueError
        If ts is not a list or np.array.
        If query is not a list or np.array.
        If ts or query is not one dimensional.
    """
    ts, query = _precheck_series_and_query(ts, query)

    if normalize_query:
        query = (query - np.mean(query)) / np.std(query)
        
    n = len(ts)
    m = len(query)
    x = np.append(ts, np.zeros([1, n]))
    y = np.append(np.flipud(query), np.zeros([1, 2 * n - m]))
    
    X = np.fft.fft(x)
    Y = np.fft.fft(y)
    Y.resize(X.shape)
    Z = X * Y
    z = np.fft.ifft(Z)
    
    sumy = np.sum(y)
    sumy2 = np.sum(y ** 2)
    
    cum_sumx = np.cumsum(x)
    cum_sumx2 = np.cumsum(x ** 2)
    
    sumx2 = cum_sumx2[m:n] - cum_sumx2[0:n-m]
    sumx = cum_sumx[m:n] - cum_sumx[0:n-m]
    meanx = sumx / m
    sigmax2 = (sumx2 / m) - (meanx**2)
    sigmax = np.sqrt(sigmax2)
    
    dist = (sumx2 - 2 * sumx * meanx + m * (meanx ** 2)) \
        / sigmax2 - 2 * (z[m:n] - sumy * meanx) \
        / sigmax + sumy2
    dist = np.absolute(np.sqrt(dist))
    
    if corr_coef:
        return 1 - np.absolute(dist) / (2 * m)
    
    return dist


def mass2(ts, query):
    """
    Compute the distance profile for the given query over the given time 
    series. Optionally, the correlation coefficient can be returned.

    Parameters
    ----------
    ts : array_like
        The array to create a rolling window on.
    query : array_like
        The query.

    Returns
    -------
    An array of distances.

    Raises
    ------
    ValueError
        If ts is not a list or np.array.
        If query is not a list or np.array.
        If ts or query is not one dimensional.
    """
    ts, query = _precheck_series_and_query(ts, query)

    n = len(ts)
    m = len(query)
    x = ts
    y = query

    meany = np.mean(y)
    sigmay = np.std(y)
    
    meanx = _moving_average(x, m)
    meanx = np.append(np.ones([1, len(x) - len(meanx)]), meanx)
    sigmax = _moving_std(x, m)
    sigmax = np.append(np.zeros([1, len(x) - len(sigmax)]), sigmax)
    
    y = np.append(np.flip(y), np.zeros([1, n - m]))
    
    X = np.fft.fft(x)
    Y = np.fft.fft(y)
    Y.resize(X.shape)
    Z = X * Y
    z = np.fft.ifft(Z)
    
    dist = 2 * (m - (z[m - 1:n] - m * meanx[m - 1:n] * meany) / 
                    (sigmax[m - 1:n] * sigmay))
    dist = np.sqrt(dist)
    
    return dist


def mass3(ts, query, pieces):
    """
    Compute the distance profile for the given query over the given time 
    series. This version of MASS is hardware efficient given the right number
    of pieces.

    Parameters
    ----------
    ts : array_like
        The array to create a rolling window on.
    query : array_like
        The query.
    pieces : int
        Number of pieces to process. This is best as a power of 2.

    Returns
    -------
    An array of distances.

    Raises
    ------
    ValueError
        If ts is not a list or np.array.
        If query is not a list or np.array.
        If ts or query is not one dimensional.
        If pieces is less than the length of the query.
    """
    ts, query = _precheck_series_and_query(ts, query)

    m = len(query)
    
    if pieces < m:
        raise ValueError('pieces should be larger than the query length.')
    
    n = len(ts)
    k = pieces
    x = ts
    dist = np.array([])
    
    # compute stats in O(n)
    meany = np.mean(query)
    sigmay = np.std(query)
    
    meanx = _moving_average(x, m)
    meanx = np.append(np.ones([1, len(x) - len(meanx)]), meanx)
    sigmax = _moving_std(x, m)
    sigmax = np.append(np.zeros([1, len(x) - len(sigmax)]), sigmax)
    
    # reverse the query and append zeros
    y = np.append(np.flip(query), np.zeros(pieces - m))
    
    step_size = k - m + 1
    stop = n - k + 1
       
    for j in range(0, stop, step_size):
        # The main trick of getting dot products in O(n log n) time
        X = np.fft.fft(x[j:j + k])
        Y = np.fft.fft(y)
        
        Z = X * Y
        z = np.fft.ifft(Z)
            
        d = 2 * (m-(z[m - 1:k] - m * meanx[m + j - 1:j + k] * meany) /
                   (sigmax[m + j - 1:j + k] * sigmay))
        d = np.sqrt(d)
        dist = np.append(dist, d)
   
    j = j + k - m
    k = n - j - 1
    if k >= m:
        X = np.fft.fft(x[j:n-1])
        y = y[0:k]

        Y = np.fft.fft(y)
        Z = X * Y
        z = np.fft.ifft(Z)

        d = 2 * (m-(z[m - 1:k] - m * meanx[j + m - 1:n - 1] * meany) /
                 (sigmax[j + m - 1:n - 1] * sigmay))
       
        d = np.sqrt(d)
        dist = np.append(dist, d)
    
    return np.array(dist)


# -*- coding: utf-8 -*-

"""
MASS (Mueen's Algorithm for Similarity Search)
==============================================

Provides
	1. MASS - the original implementation
	2. MASS2 - a quicker version of MASS
	3. MASS3 - a piecewise version of MASS2

Example Usage
-------------
>>> import mass_ts
>>> import numpy as np
>>> ts = np.loadtxt('ts.txt')
>>> query = np.loadtxt('query.txt')
>>> distance = mass_ts.mass(ts, query)

Citations
---------
Abdullah Mueen, Yan Zhu, Michael Yeh, Kaveh Kamgar, Krishnamurthy Viswanathan, Chetan Kumar Gupta and Eamonn Keogh (2015), The Fastest Similarity Search Algorithm for Time Series Subsequences under Euclidean Distance, URL: http://www.cs.unm.edu/~mueen/FastestSimilaritySearch.html
"""

__author__ = """Tyler Marrs"""
__email__ = 'tylerwmarrs@gmail.com'
__version__ = '0.1.1'


from mass_ts._mass_ts import mass, mass2, mass3


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

"""Tests for `mass_ts` package."""

import os

import pytest

import numpy as np

import mass_ts
from mass_ts import _mass_ts as mts

MODULE_PATH = mass_ts.__path__[0]


def test_is_array_like_invalid():
    assert(mts._is_array_like(1) == False)
    assert(mts._is_array_like('adf') == False)
    assert(mts._is_array_like({'a': 1}) == False)
    assert(mts._is_array_like(set([1, 2, 3])) == False)


def test_is_array_like_valid():
    assert(mts._is_array_like(np.array([1])) == True)
    assert(mts._is_array_like([1, ]) == True)
    assert(mts._is_array_like((1, 2,)) == True)


def test_is_one_dimensional_invalid():
    a = np.array([
        [1, 2, 3],
        [2, 3, 4],
        [3, 4, 5],
        [4, 5, 6]
    ])
    assert(mts._is_one_dimensional(a) == False)


def test_is_one_dimensional_valid():
    a = np.array([1, 2, 3, 4])
    assert(mts._is_one_dimensional(a) == True)


def test_to_np_array_exception():
    with pytest.raises(ValueError) as excinfo:
        mts._to_np_array('s')
        assert 'Unable to convert to np.ndarray!' in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        mts._to_np_array(1)
        assert 'Unable to convert to np.ndarray!' in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        mts._to_np_array(set([1, 2, 3]))
        assert 'Unable to convert to np.ndarray!' in str(excinfo.value)


def test_to_np_array_valid():
    actual = mts._to_np_array([1, 2, 3])
    desired = np.array([1, 2, 3])
    np.testing.assert_equal(actual, desired)

    actual = mts._to_np_array((1, 2, 3,))
    desired = np.array([1, 2, 3])
    np.testing.assert_equal(actual, desired)

    actual = mts._to_np_array(np.array([1, 2, 3]))
    desired = np.array([1, 2, 3])
    np.testing.assert_equal(actual, desired)


def test_precheck_series_and_query_valid():
    ts = [1, 2, 3, 4, 5, 6, 7, 8]
    q = [1, 2, 3, 4]

    actual_ts, actual_q = mts._precheck_series_and_query(ts, q)
    np.testing.assert_equal(actual_ts, np.array(ts))
    np.testing.assert_equal(actual_q, np.array(q))


def test_precheck_series_and_query_invalid():
    with pytest.raises(ValueError) as excinfo:
        mts._precheck_series_and_query('1', [1, 2, 3])
        assert 'Invalid ts value given. Must be array_like!' \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        mts._precheck_series_and_query([1, 2, 3], '1')
        assert 'Invalid query value given. Must be array_like!' \
            in str(excinfo.value)


def test_rolling_window():
    a = np.array([1, 2, 3, 4, 5, 6])
    actual = mts._rolling_window(a, 3)
    desired = np.array([
        [1, 2, 3],
        [2, 3, 4],
        [3, 4, 5],
        [4, 5, 6]
    ])

    np.testing.assert_equal(actual, desired)


def test_moving_average():
    a = np.array([1, 2, 3, 4, 5, 6])
    actual = mts._moving_average(a, 3)
    desired = np.array([2., 3., 4., 5.])

    np.testing.assert_equal(actual, desired)


def test_moving_std():
    a = np.array([1, 2, 3, 4, 5, 6])
    actual = mts._moving_std(a, 3)
    desired = np.array([0.81649658, 0.81649658, 0.81649658, 0.81649658])

    np.testing.assert_almost_equal(actual, desired)


def test_mass():
    ts = np.array([1, 1, 1, 2, 1, 1, 4, 5])
    query = np.array([2, 1, 1, 4])
    actual = mts.mass(ts, query)
    desired = np.array([
        3.43092352e+00, 3.43092352e+00, 2.98023224e-08, 1.85113597e+00
    ])
    
    np.testing.assert_almost_equal(actual, desired)


def test_mass_corr_coef():
    ts = np.array([1, 1, 1, 2, 1, 1, 4, 5])
    query = np.array([2, 1, 1, 4])
    actual = mts.mass(ts, query, corr_coef=True)
    desired = np.array([0.57113456, 0.57113456, 1., 0.768608])
    
    np.testing.assert_almost_equal(actual, desired)


def test_mass2():
    ts = np.array([1, 1, 1, 2, 1, 1, 4, 5])
    query = np.array([2, 1, 1, 4])
    actual = mts.mass2(ts, query)
    desired = np.array([
        0.67640791-1.37044402e-16j,
        3.43092352+0.00000000e+00j,
        3.43092352+1.02889035e-17j,
        0.+0.00000000e+00j,
        1.85113597+1.21452707e-17j
    ])
    
    np.testing.assert_almost_equal(actual, desired)


def test_mass3():
    ts = np.array([1, 1, 1, 2, 1, 1, 4, 5])
    query = np.array([2, 1, 1, 4])
    pieces = 8
    distances = mts.mass3(ts, query, pieces)
    desired = np.array([
        0.67640791,
        3.43092352,
        3.43092352,
        0.,
        1.85113597
    ])

    np.testing.assert_almost_equal(distances, desired)


def test_mass_robotdog():
    """Sanity check that compares results from UCR use case."""
    robot_dog = np.loadtxt(
        os.path.join(MODULE_PATH, '..', 'tests', 'robot_dog.txt'))
    carpet_walk = np.loadtxt(
        os.path.join(MODULE_PATH, '..', 'tests', 'carpet_query.txt'))

    distances = mts.mass(robot_dog, carpet_walk)
    min_idx = np.argmin(distances) + 1

    assert(min_idx == 7479)


def test_mass2_robotdog():
    """Sanity check that compares results from UCR use case."""
    robot_dog = np.loadtxt(
        os.path.join(MODULE_PATH, '..', 'tests', 'robot_dog.txt'))
    carpet_walk = np.loadtxt(
        os.path.join(MODULE_PATH, '..', 'tests', 'carpet_query.txt'))

    distances = mts.mass2(robot_dog, carpet_walk)
    min_idx = np.argmin(distances)

    assert(min_idx == 7479)


def test_mass3_robotdog():
    """Sanity check that compares results from UCR use case."""
    robot_dog = np.loadtxt(
        os.path.join(MODULE_PATH, '..', 'tests', 'robot_dog.txt'))
    carpet_walk = np.loadtxt(
        os.path.join(MODULE_PATH, '..', 'tests', 'carpet_query.txt'))

    distances = mts.mass3(robot_dog, carpet_walk, 256)
    min_idx = np.argmin(distances)

    assert(min_idx == 7479)